/*-
 * Copyright (c) 2004 B. Luevelsmeyer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/* Measure temperature from a Voltcraft 304 4-channel
 * thermometer.
 *
 * The thermometer has 4 sensors (one per channel). It is
 * connected to the computer via a serial line. Upon receiving a
 * byte "A", it sends the 4 values to the computer. The protocol
 * is described in more detail later in this source.
 *
 * Call with the serial port (e.g. "/dev/cuaa1") as command line
 * argument.
 *
 * The output is the temperature in tenths of degree Celsius, one
 * line per channel; channels without a sensor connected are
 * output as "---". Example:
 *      193
 *      253
 *      ---
 *      ---
 */

#include <stdio.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>



int main(int argc, char **argv)
{
    int sl;     /* the serial line */

    if(argc!=2)
    {
        fputs("usage: vc304 serial_line\n",stderr);
        return -1;
    }

    /* open serial line */
    if((sl = open(argv[1],O_RDWR)) == -1)
    {
        perror("vc304: open() failed");
        return -2;
    }

    /* setup serial line for themometer use:
     * 9600 baud, 8 data bits, no parity, 1 stop bit, no flow
     * control
     */
    {
        struct termios tp;
        if(tcgetattr(sl,&tp))
        {
            perror("vc304: tcgetattr() failed");
            return -3;
        }
        cfmakeraw(&tp);             /* no conversions */
        cfsetspeed(&tp,9600);       /* 9600 baud */
        tp.c_cflag &= ~CSIZE;       /* clear bitsize */
        tp.c_cflag |= CS8;          /* 8 data bits */
        tp.c_cflag &= ~PARENB;      /* no parity */
        tp.c_cflag &= ~CSTOPB;      /* 1 stop bit */
        tp.c_cflag |= CLOCAL;
        tp.c_cflag &= ~CRTSCTS;     /* no RTS/CTS flow control */
        tp.c_iflag &= ~(IXON|IXOFF);/* no Xon/Xoff flow control */
        if(tcsetattr(sl,TCSANOW,&tp))
        {
            perror("vc304: tcsetattr() failed");
            return -4;
        }
    }

    /* take a measurement */
    {
            /* Protocol:
             * To trigger a measurement, send a single byte "A".
             * The thermometer sends 45 bytes as a reply.
             *
             * byte | meaning
             * -----+--------
             *   0   unknown
             *   1   0x80 set for "Celsius", else "Farenheit"
             *       0x02 set for "max/min"
             *       0x20 set for "hold"
             *       0x10 set for "delta-rel"
             *       0x08 set for "T1-T2"
             *   2   unknown
             *   3   state channel 1:
             *       0x1 - ok, Celsius
             *       0x2 - not connected, Celsius
             *       0x4 - ok, Farenheit
             *       0x10 - not connected, Farenheit
             *   4   state channel 2, see 3
             *   5   state channel 3, see 3
             *   6   state channel 4, see 3
             *   7   value channel 1 upper
             *   8   value channel 1 lower
             *   9   value channel 2 upper
             *  10   value channel 2 lower
             *  11   value channel 3 upper
             *  12   value channel 3 lower
             *  13   value channel 4 upper
             *  14   value channel 4 lower
             *  rest unknown
             *
             * To get a temperature in Celsius, send "A" and
             * receive 45 Bytes. Check that byte 1 is 0x80 and
             * that your channel has a state of 0x1. Shift the
             * upper value 8 bits to the left and or the lower
             * value to it; the result is a 2's complement
             * 16-bit-value in tenths of Celsius. Divide by 10 to
             * get Celsius.
             *
             * This information is from reverse engineering, so
             * it is incomplete and may contain errors.
             */
        struct
        {
            unsigned char unknown1;
            unsigned char flags;        /* should be 0x80 */
            unsigned char unknown2;
            unsigned char state[4];     /* 0x1, otherwise "not connected" */
            unsigned char value[4][2];  /* Motorola-ordered 16-bit-values */
            unsigned char unknown3[30];
        } result;
        /* to check that the struct is unpadded */
        char dummy[((sizeof result)==45)*2-1];

        int retb;       /* number of read bytes */
        int chan;       /* the current channel */

        /* trigger measurement */
        if(write(sl,"A",1)!=1)
        {
            perror("vc304: write() failed");
            return -5;
        }

        /* receive answer */
        for(retb=0;retb<(int)sizeof result;)
        {
            int bytes = read(sl,(char*)&result+retb,sizeof result - retb);
            if(bytes<=0)
            {
                perror("vc304: read() failed");
                return -6;
            }
            retb += bytes;
        }

        /* check the thermometer is set up sensibly */
        if(result.flags!=0x80)
        {
            fputs("vc304: thermometer settings bad\n",stderr);
            return -7;
        }

        /* output results */
        for(chan=0;chan<4;chan++)
            if(result.state[chan]!=0x1)
                puts("---");
            else
            {
                int val = result.value[chan][0]*0x100+result.value[chan][1];
                if(val&1<<15) /* negative -> get 2's complement and negate it */
                    val = -((val^0xffff)+1);
                printf("%d\n",val);
            }
    }

    if(close(sl))
    {
        perror("vc304: close() failed");
        return -8;
    }

    return 0;
}
